print("import: 'pyutilib'")
import pyutilib

print("import: 'pyutilib.autotest'")
import pyutilib.autotest

print("import: 'pyutilib.autotest.tests'")
import pyutilib.autotest.tests

print("import: 'pyutilib.common'")
import pyutilib.common

print("import: 'pyutilib.component'")
import pyutilib.component

print("import: 'pyutilib.component.app'")
import pyutilib.component.app

print("import: 'pyutilib.component.app.tests'")
import pyutilib.component.app.tests

print("import: 'pyutilib.component.config'")
import pyutilib.component.config

print("import: 'pyutilib.component.config.tests'")
import pyutilib.component.config.tests

print("import: 'pyutilib.component.core'")
import pyutilib.component.core

print("import: 'pyutilib.component.core.tests'")
import pyutilib.component.core.tests

print("import: 'pyutilib.component.executables'")
import pyutilib.component.executables

print("import: 'pyutilib.component.loader'")
import pyutilib.component.loader

print("import: 'pyutilib.component.loader.tests'")
import pyutilib.component.loader.tests

print("import: 'pyutilib.dev'")
import pyutilib.dev

print("import: 'pyutilib.dev.tests'")
import pyutilib.dev.tests

print("import: 'pyutilib.enum'")
import pyutilib.enum

print("import: 'pyutilib.enum.tests'")
import pyutilib.enum.tests

print("import: 'pyutilib.excel'")
import pyutilib.excel

print("import: 'pyutilib.excel.tests'")
import pyutilib.excel.tests

print("import: 'pyutilib.math'")
import pyutilib.math

print("import: 'pyutilib.math.tests'")
import pyutilib.math.tests

print("import: 'pyutilib.misc'")
import pyutilib.misc

print("import: 'pyutilib.misc.tests'")
import pyutilib.misc.tests

print("import: 'pyutilib.misc.tests.import_data'")
import pyutilib.misc.tests.import_data

print("import: 'pyutilib.misc.tests.import_data.a'")
import pyutilib.misc.tests.import_data.a

print("import: 'pyutilib.misc.tests.import_data.b'")
import pyutilib.misc.tests.import_data.b

print("import: 'pyutilib.ply'")
import pyutilib.ply

print("import: 'pyutilib.pyro'")
import pyutilib.pyro

print("import: 'pyutilib.services'")
import pyutilib.services

print("import: 'pyutilib.subprocess'")
import pyutilib.subprocess

print("import: 'pyutilib.subprocess.tests'")
import pyutilib.subprocess.tests

print("import: 'pyutilib.th'")
import pyutilib.th

print("import: 'pyutilib.th.tests'")
import pyutilib.th.tests

print("import: 'pyutilib.workflow'")
import pyutilib.workflow

print("import: 'pyutilib.workflow.tests'")
import pyutilib.workflow.tests

